import vii
from fwk.wutils import parseargs
from fwk.coloring import ccolors
import numpy as np

def initBL(Re, Minf, CFL0, nSections, xtrF=[-1, -1], span=0, verb=None):
    """ Initialize boundary layer solver.
    
    Params
    ------
    - Re: Flow Reynolds number.
    - Minf: Freestream Mach number.
    - CFL0: Initial CFL number for time integration.
    - nSections: Number of sections in the domain.
    - xtrF: Forced transition location [upper, lower].
    - span: Wing span (not used for 2D computations.
    - verb: Verbosity level of the solver.
    
    Return
    ------
    - solver: vii::Driver class.
    """
    if Re<=0.:
        print(ccolors.ANSI_RED, "dart::vUtils Error : Negative Reynolds number.", Re, ccolors.ANSI_RESET)
        raise RuntimeError("Invalid parameter")
    if Minf<0.:
        print(ccolors.ANSI_RED, "dart::vUtils Error : Negative Mach number.", Minf, ccolors.ANSI_RESET)
        raise RuntimeError("Invalid parameter")
    elif Minf>=1.:
        print(ccolors.ANSI_YELLOW, "dart::vUtils Warning : (Super)sonic freestream Mach number.", Minf, ccolors.ANSI_RESET)
    if nSections < 0:
        print(ccolors.ANSI_RED, "dart::vUtils Fatal error : Negative number of sections.", nSections, ccolors.ANSI_RESET)
        raise RuntimeError("Invalid parameter")
    if verb is None:
      args = parseargs()
      _verbose = args.verb
    else:
      if not(0<=verb<=3):
        print(ccolors.ANSI_RED, "dart::vUtils Fatal error : verbose not in valid range.", _verbose, ccolors.ANSI_RESET)
        raise RuntimeError("Invalid parameter")
      else:
        _verbose = verb
    _xtrF = [-1, -1]
    for i, xtr in enumerate(xtrF):
            if not(0 <= xtr <= 1) and xtr != -1:
                raise RuntimeError("Incorrect forced transition location.")
            _xtrF[i] = xtr
    solver = vii.Driver(Re, Minf, CFL0, nSections, xtrF[0], xtrF[1], span, verbose=_verbose)
    return solver

def initDart(pbl, blw, vSolver, iters = 25):
    """Initialize Newton solver and create the interface object
       between dart and the viscous solver.

    Parameters
    ----------
    pbl (dart.Problem object): Problem formulation.
    blw (np.ndarray(dart.Blowing, dart.Blowing)): Blowing boundaries for (airfoil, wake).
    vSolver (vii.Driver object): Viscous solver.

    Return
    ------
    solverAPI (DartInterface object): Interface between Dart and the viscous solver.
    """
    import dart
    from vii.interfaces.dart.DartInterface import DartInterface as dInterface
    from tbox.solvers import LinearSolver

    args = parseargs()
    dartSolver = dart.Newton(pbl, LinearSolver().pardiso(), nthrds=args.k, vrb=args.verb, mIt=iters)
    solverAPI = dInterface(dartSolver, vSolver, blw)
    return solverAPI

def mesh(file, pars):
    """Initialize mesh and mesh writer

    Parameters
    ----------
    file : str
        file contaning mesh (w/o extention)
    pars : dict
        parameters for mesh
    """
    import tbox.gmsh as tmsh
    # Load the mesh.
    msh = tmsh.MeshLoader(file,__file__).execute(**pars)
    return msh


def getSolution(vSolver, iSec=0):
    """Extract viscous solution.
    Args
    ----
    - vSolver: vii::Driver class. Drive of the viscous calculations
    - iSec (int): Marker of the section (default: 0)
    """
    if iSec<0:
        raise RuntimeError("dart::viscU Invalid section number", iSec)
    
    solverOutput = vSolver.getSolution(iSec)

    sln = {'theta'    : solverOutput[0],
            'H'        : solverOutput[1],
            'N'        : solverOutput[2],
            'ue'       : solverOutput[3],
            'Ct'       : solverOutput[4],
            'deltaStar': solverOutput[5],
            'Ret'      : solverOutput[6],
            'Cd'       : solverOutput[7],
            'Cf'       : np.asarray(solverOutput[8])*np.asarray(solverOutput[3])**2,
            'Hstar'    : solverOutput[9],
            'Hstar2'   : solverOutput[10],
            'Hk'       : solverOutput[11],
            'Cteq'     : solverOutput[12],
            'Us'       : solverOutput[13],
            'delta'    : solverOutput[14],
            'x/c'      : solverOutput[15],
            'xtrT'     : vSolver.getxtr(iSec, 0),
            'xtrB'     : vSolver.getxtr(iSec, 1),
            'Cdt'      : vSolver.Cdt,
            'Cdf'      : vSolver.Cdf,
            'Cdp'      : vSolver.Cdp}
    return sln

def writeFile(wData, Re, toW=['deltaStar', 'H', 'Cf', 'Ct', 'ue', 'delta']):
    """Write the results in airfoil_viscous.dat
    """
    # Write
    print('Writing file: airfoil_viscous.dat...', end = '')
    f = open('airfoil_viscous.dat', 'w+')

    f.write('$Aerodynamic coefficients\n')
    f.write('             Re              Cd             Cdp             Cdf         xtr_top         xtr_bot\n')
    f.write('{0:15.6f} {1:15.6f} {2:15.6f} {3:15.6f} {4:15.6f} {5:15.6f}\n'.format(Re, wData['Cdt'], wData['Cdp'],wData['Cdf'], wData['xtrT'], wData['xtrB']))
    f.write('$Boundary layer variables\n')
    f.write('            x/c')

    for s in toW:
        f.write(' {0:>15s}'.format(s))
    f.write('\n')

    for i in range(len(wData['x/c'])):
        f.write('{0:15.6f}'.format(wData['x/c'][i]))
        for s in toW:
            f.write(' {0:15.6f}'.format(wData[s][i]))
        f.write('\n')

    f.close()
    print('done.')